/* This file was generated by plugin 'Nordic Semiconductor nRF5x v.1.2.2' (BDS version 1.1.3139.0) */

#include "ble_quickdraw.h"
#include <string.h>
#include "nordic_common.h"
#include "ble_srv_common.h"
#include "app_util.h"
#include "app_util_bds.h"

#define OPCODE_LENGTH 1 /**< Length of opcode inside quickdraw packet. */
#define HANDLE_LENGTH 2 /**< Length of handle inside quickdraw packet. */

/* TODO Consider changing the max values if encoded data for characteristic/descriptor is fixed length */ 
#define MAX_XYZ_LEN (BLE_L2CAP_MTU_DEF - OPCODE_LENGTH - HANDLE_LENGTH) /**< Maximum size of a transmitted xyz. */ 

/**@brief Function for encoding xyz.
 *
 * @param[in]   p_xyz              xyz characteristic structure to be encoded.
 * @param[out]  p_encoded_buffer   Buffer where the encoded data will be written.
 *
 * @return      Size of encoded data.
 */
static uint8_t xyz_encode(ble_quickdraw_xyz_t * p_xyz, uint8_t * encoded_buffer)
{
    uint8_t len = 0; 
    len += bds_int16_encode(&p_xyz->x, &encoded_buffer[len]); 
    len += bds_int16_encode(&p_xyz->y, &encoded_buffer[len]); 
    len += bds_int16_encode(&p_xyz->z, &encoded_buffer[len]); 
    return len;
}


/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_quickdraw       quickdraw Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_quickdraw_t * p_quickdraw, ble_evt_t * p_ble_evt)
{
    p_quickdraw->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_quickdraw       quickdraw Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_quickdraw_t * p_quickdraw, ble_evt_t * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_quickdraw->conn_handle = BLE_CONN_HANDLE_INVALID;
}

/**@brief Function for handling the Write event.
 *
 * @param[in]   p_quickdraw       quickdraw Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_write(ble_quickdraw_t * p_quickdraw, ble_gatts_evt_write_t * p_ble_evt)
{
    
    if(p_ble_evt->handle == p_quickdraw->xyz_handles.cccd_handle)
    {
        if(p_quickdraw->evt_handler != NULL)
        {
            ble_quickdraw_evt_t evt;
            evt.evt_type = BLE_QUICKDRAW_XYZ_EVT_CCCD_WRITE;
            bds_uint16_decode(p_ble_evt->len, p_ble_evt->data, &evt.params.cccd_value);
            p_quickdraw->evt_handler(p_quickdraw, &evt);
        }
    } 
}

/**@brief Authorize WRITE request event handler.
 *
 * @details Handles WRITE events from the BLE stack.
 *
 * @param[in]   p_sc_ctrlpt  SC Ctrlpt structure.
 * @param[in]   p_gatts_evt  GATTS Event received from the BLE stack.
 *
 */
static void on_rw_authorize_request(ble_quickdraw_t * p_quickdraw, ble_gatts_evt_t * p_gatts_evt)
{
    ble_gatts_evt_rw_authorize_request_t * p_auth_req = &p_gatts_evt->params.authorize_request;
    if (p_auth_req->type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
    {
        if (   (p_gatts_evt->params.authorize_request.request.write.op
                != BLE_GATTS_OP_PREP_WRITE_REQ)
            && (p_gatts_evt->params.authorize_request.request.write.op
                != BLE_GATTS_OP_EXEC_WRITE_REQ_NOW)
            && (p_gatts_evt->params.authorize_request.request.write.op
                != BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL)
           )
        {
        
        }
    }
}

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_quickdraw       quickdraw Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
void ble_quickdraw_on_ble_evt(ble_quickdraw_t * p_quickdraw, ble_evt_t * p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_quickdraw, p_ble_evt);
            break;
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_quickdraw, p_ble_evt);
            break;
        case BLE_GATTS_EVT_WRITE:
            on_write(p_quickdraw, &p_ble_evt->evt.gatts_evt.params.write);
            break;
         case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
            on_rw_authorize_request(p_quickdraw, &p_ble_evt->evt.gatts_evt);
            break;
        default:
            //No implementation needed.
            break;
    }
}

/**@brief Function for initializing the quickdraw. */
uint32_t ble_quickdraw_init(ble_quickdraw_t * p_quickdraw, const ble_quickdraw_init_t * p_quickdraw_init)
{
    uint32_t err_code;
    ble_uuid_t ble_uuid;

    // Initialize service structure
    p_quickdraw->evt_handler = p_quickdraw_init->evt_handler;
    p_quickdraw->conn_handle = BLE_CONN_HANDLE_INVALID;
    
    // Add a custom base UUID.
    ble_uuid128_t bds_base_uuid = {{0x3A, 0x79, 0x82, 0x2D, 0xB4, 0xE4, 0x42, 0x8D, 0xFD, 0x4B, 0x76, 0x44, 0x00, 0x00, 0x3E, 0x0A}};
    uint8_t       uuid_type;
    err_code = sd_ble_uuid_vs_add(&bds_base_uuid, &uuid_type);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    }
    ble_uuid.type = uuid_type;
    ble_uuid.uuid = 0x2BC7;
        
    // Add service
    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_quickdraw->service_handle);
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    } 

    // Add xyz characteristic
    ble_quickdraw_xyz_t xyz_initial_value = p_quickdraw_init->ble_quickdraw_xyz_initial_value; 
    p_quickdraw->is_xyz_read_supported = p_quickdraw_init->is_xyz_read_supported;

    uint8_t xyz_encoded_value[MAX_XYZ_LEN];
    ble_add_char_params_t add_xyz_params;
    memset(&add_xyz_params, 0, sizeof(add_xyz_params));
    
    add_xyz_params.uuid                = 0x4B9E;
    add_xyz_params.uuid_type           = ble_uuid.type; 
    add_xyz_params.max_len             = MAX_XYZ_LEN;
    add_xyz_params.init_len            = xyz_encode(&xyz_initial_value, xyz_encoded_value);
    add_xyz_params.p_init_value        = xyz_encoded_value; 
    add_xyz_params.char_props.notify   = 1; 
    add_xyz_params.char_props.read = p_quickdraw->is_xyz_read_supported; 
    add_xyz_params.read_access         = SEC_OPEN; 
    add_xyz_params.cccd_write_access   = SEC_OPEN;
    // 1 for variable length and 0 for fixed length.
    add_xyz_params.is_var_len          = 1; 

    err_code = characteristic_add(p_quickdraw->service_handle, &add_xyz_params, &(p_quickdraw->xyz_handles));
    if (err_code != NRF_SUCCESS)
    {
        return err_code;
    } 

    return NRF_SUCCESS;
}

/**@brief Function for setting the xyz. */
uint32_t ble_quickdraw_xyz_set(ble_quickdraw_t * p_quickdraw, ble_quickdraw_xyz_t * p_xyz)
{
    ble_gatts_value_t gatts_value;
    uint8_t encoded_value[MAX_XYZ_LEN];

    // Initialize value struct.
    memset(&gatts_value, 0, sizeof(gatts_value));

    gatts_value.len     = xyz_encode(p_xyz, encoded_value);
    gatts_value.offset  = 0;
    gatts_value.p_value = encoded_value;

    return sd_ble_gatts_value_set(p_quickdraw->conn_handle, p_quickdraw->xyz_handles.value_handle, &gatts_value);
}

/**@brief Function for sending the xyz. */
uint32_t ble_quickdraw_xyz_send(ble_quickdraw_t * p_quickdraw, ble_quickdraw_xyz_t * p_xyz)
{
    uint32_t err_code = NRF_SUCCESS;

    if (p_quickdraw->conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        ble_gatts_hvx_params_t hvx_params;
        uint8_t encoded_value[MAX_XYZ_LEN];
        uint16_t hvx_len;

        // Initialize value struct.
        memset(&hvx_params, 0, sizeof(hvx_params));

        hvx_len           = xyz_encode(p_xyz, encoded_value);
        hvx_params.handle = p_quickdraw->xyz_handles.value_handle;
        hvx_params.type   = BLE_GATT_HVX_NOTIFICATION;
        hvx_params.p_len  = &hvx_len;
        hvx_params.offset = 0;
        hvx_params.p_data = encoded_value;

        err_code = sd_ble_gatts_hvx(p_quickdraw->conn_handle, &hvx_params);
    }
    else
    {
        err_code = NRF_ERROR_INVALID_STATE;
    }

    return err_code;
}

